import dominate
from pathlib import Path
from typing import Any, Dict, List, Optional
from ._utils import (
    _map_speaker_names,
    _get_segement_speaker,
    _group_transcript_segements_by_speaker,
)


test = """function play(t) {
var audio = document.getElementById('audio');
audio.currentTime = t;
audio.play();
}"""


def generate(
    transcript: dict,
    diarisation: Optional[List[Dict[str, Any]]] = None,
    audio_file: Optional[Path] = None,
    speaker_mapping: Optional[dict] = None,
):
    """Create html page from conversation.

    Parameters
    ----------
    transcript : dict
        Transcript from transcribe module.
    diarisation : dict
        Speaker diarisation from from diarisation module.
    audio_file : PosixPath | None
        Path to audio file. If provided, audio will be embedded in the
        report and timestamps will link to audio times.
    speaker_mapping : dict
        Mapping of old speaker names to new speaker names.
        For example, {"0": "Alice", "1": "Bob"}.

    Returns
    -------
    report : dominate.document
        Conversation report as a dominate document.
    """
    if diarisation is not None:
        transcript = _group_transcript_segements_by_speaker(transcript, diarisation)

    if speaker_mapping is not None:
        transcript = _map_speaker_names(transcript, speaker_mapping)

    doc = dominate.document(title="Conversation")

    with doc.head:
        dominate.tags.script(src="https://cdn.tailwindcss.com")

    with doc:
        with dominate.tags.div(cls="ml-12 mr-12 mt-6"):
            if audio_file is not None:
                with dominate.tags.audio(
                    id="audio", controls="controls", cls="w-full border"
                ):
                    dominate.tags.source(src=str(audio_file), type="audio/wav")

                dominate.tags.script(test)

            with dominate.tags.div(id="conversation", cls="pt-2"):
                previous_speaker = None
                for seg in transcript["segments"]:
                    start = seg["start"]
                    stop = seg["end"]
                    speaker = _get_segement_speaker(seg)

                    with dominate.tags.p(cls="mr-2 m-1 flex"):
                        if diarisation is not None:
                            if speaker != previous_speaker:
                                dominate.tags.input_(
                                    value=f"{speaker}",
                                    cls="border m-1 bg-blue-50 mb-2 mt-4 p-1 pl-2 pr-2",
                                    type="button",
                                    onclick=f"play({start})",
                                )
                                previous_speaker = speaker

                        dominate.tags.p(
                            f"{seg['text']}",
                            cls="ml-8 mr-8 m-1",
                            onclick=f"play({start})",
                        )

    return doc



from conversations import Conversation, load_conversation
from conversations._conversations import _create_default_save_filename
from pathlib import Path
import pooch
import dominate
import os
import tempfile
import pytest
import pickle
from datetime import datetime, timezone, timedelta


audio_file = Path(
    pooch.retrieve(
        url="https://project-test-data-public.s3.amazonaws.com/test_audio.m4a",
        known_hash="md5:77d8b60c54dffbb74d48c4a65cd59591",
    )
)


def test_conv_type():
    conv = Conversation(recording=audio_file)
    assert isinstance(conv, Conversation)


def test_conv_report():
    conv = Conversation(recording=audio_file)
    conv.transcribe()
    html_report = conv.report()
    assert isinstance(html_report, dominate.document)


def test_save_and_load():
    # Create a temporary directory to store the conversation file
    with tempfile.TemporaryDirectory() as temp_dir:
        conversation_file = Path(temp_dir) / "conversation.pkl"

        # Create a Conversation object, transcribe, and diarise
        conversation = Conversation(recording=Path(audio_file))
        conversation.transcribe()
        conversation.diarise()

        # Save the Conversation object to disk
        conversation.save(conversation_file)

        # Load the Conversation object from disk
        loaded_conversation = load_conversation(conversation_file)

        # Check if the loaded Conversation object has the same attributes as the original
        assert loaded_conversation._recording == conversation._recording
        assert loaded_conversation._num_speakers == conversation._num_speakers
        assert loaded_conversation._transcription == conversation._transcription
        assert loaded_conversation._diarisation == conversation._diarisation

        # Remove the temporary file
        os.remove(conversation_file)


def test_invalid_conversation():
    # Create a temporary directory to store the invalid conversation file
    with tempfile.TemporaryDirectory() as temp_dir:
        invalid_conversation_file = Path(temp_dir) / "invalid_conversation.pkl"

        # Save a non-Conversation object to disk
        with open(invalid_conversation_file, "wb") as file:
            pickle.dump("not a conversation object", file)

        # Try to load the invalid object and check if a ValueError is raised
        with pytest.raises(
            ValueError,
            match="The loaded object is not an instance of the Conversation class.",
        ):
            load_conversation(invalid_conversation_file)

        # Remove the temporary file
        os.remove(invalid_conversation_file)


@pytest.mark.parametrize(
    "audio_file_path, expected_conversation_file_path",
    [
        ("my-audio-file.mp3", "my-audio-file.conv"),
        ("my-audio-file.ext", "my-audio-file.conv"),
        ("my-audio-file.m4a", "my-audio-file.conv"),
        ("my-audio-file.wav", "my-audio-file.conv"),
        ("my-audio-file.mov", "my-audio-file.conv"),
        ("my-audio-file.mkv", "my-audio-file.conv"),
        ("my audio file.mp3", "my audio file.conv"),
        ("my_audio_file.mp3", "my_audio_file.conv"),
        ("./my-audio-file.mp3", "./my-audio-file.conv"),
        ("../../my-audio-file.mp3", "../../my-audio-file.conv"),
        ("/path/to/my/file/my-audio-file.mp3", "/path/to/my/file/my-audio-file.conv"),
        ("/path/to/my file/my-audio-file.mp3", "/path/to/my file/my-audio-file.conv"),
    ],
)
def test_create_default_filename(audio_file_path, expected_conversation_file_path):
    assert _create_default_save_filename(Path(audio_file_path)) == Path(
        expected_conversation_file_path
    )


def test_save_and_reload_conversation(tmp_path: Path):
    # Create a dummy recording file
    recording_path = audio_file
    recording_path.touch()

    # Create a new conversation
    conversation = Conversation(recording=recording_path)

    # Save the conversation with the default filename
    conversation.save()

    # Check if the saved conversation file exists
    saved_conversation_path = _create_default_save_filename(recording_path)
    assert saved_conversation_path.is_file()

    # Load the existing saved conversation with the default filename
    loaded_conversation = Conversation(recording=recording_path, reload=True)

    # Check if the loaded conversation has the same recording path as the original conversation
    assert loaded_conversation._recording == conversation._recording
    assert loaded_conversation._num_speakers == conversation._num_speakers


def test_meeting_datetime():
    # Test with the provided meeting_datetime parameter
    meeting_datetime = datetime(2022, 1, 1, 12, 0, 0, 0, tzinfo=timezone.utc)
    conv_with_datetime = Conversation(
        recording=audio_file, meeting_datetime=meeting_datetime, reload=False
    )
    assert abs(conv_with_datetime._meeting_datetime - meeting_datetime) < timedelta(
        seconds=1
    )

    # Test without the meeting_datetime parameter (get from file metadata)
    conv_without_datetime = Conversation(recording=audio_file, reload=False)
    file_metadata_datetime = conv_without_datetime._extract_datetime_from_file(
        audio_file
    )
    assert abs(
        conv_without_datetime._meeting_datetime - file_metadata_datetime
    ) < timedelta(seconds=1)


def test_attendees():
    attendees = ["Alice", "Bob", "Charlie"]
    conv_with_attendees = Conversation(
        recording=audio_file, attendees=attendees, reload=False, num_speakers=None
    )
    assert conv_with_attendees._attendees == attendees
    assert conv_with_attendees._num_speakers == len(attendees)


def test_num_speakers_and_attendees_mismatch():
    attendees = ["Alice", "Bob", "Charlie"]
    num_speakers = 2
    with pytest.raises(
        ValueError, match="The number of speakers must match the number of attendees."
    ):
        Conversation(
            recording=audio_file,
            num_speakers=num_speakers,
            attendees=attendees,
            reload=False,
        )


def test_num_speakers_and_attendees():
    attendees = ["Alice", "Bob", "Charlie"]
    num_speakers = len(attendees)
    conv_with_speakers_attendees = Conversation(
        recording=audio_file,
        num_speakers=num_speakers,
        attendees=attendees,
        reload=False,
    )
    assert conv_with_speakers_attendees._attendees == attendees
    assert conv_with_speakers_attendees._num_speakers == num_speakers


@pytest.fixture
def conversation():
    return Conversation(recording=audio_file, reload=True)


def test_summarise_new_summary(conversation):
    conversation.diarise()
    conversation.transcribe()
    summary = conversation.summarise(force=True, print_summary=False)
    assert isinstance(summary, str)
    assert len(summary) > 0


def test_summarise_existing_summary(conversation):
    conversation.diarise()
    conversation.transcribe()
    conversation._summary_automated = "Test summary"
    summary = conversation.summarise(force=False, print_summary=False)
    assert summary == "Test summary"


def test_summarise_force_new_summary(conversation):
    conversation.diarise()
    conversation.transcribe()
    conversation._summary_automated = "Test summary"
    summary = conversation.summarise(force=True, print_summary=False)
    assert summary != "Test summary"
    assert isinstance(summary, str)
    assert len(summary) > 0



from conversations.diarise import simple
from pathlib import Path
import pooch


audio_file = pooch.retrieve(
    url="https://project-test-data-public.s3.amazonaws.com/test_audio.m4a",
    known_hash="md5:77d8b60c54dffbb74d48c4a65cd59591",
)


def test_process():
    """Test whisper processing of audio."""
    segments = simple.process(audio_file=Path(audio_file))
    assert isinstance(segments, list)
    for seg in segments:
        assert "start" in seg
        assert "end" in seg
        assert "label" in seg



from pathlib import Path
from typing import Any, Dict, List, Optional
import pickle
from datetime import datetime, timezone
import os


class Conversation:
    """Conversations class.

    This class is the core of the **Conversations** package.
    Use this class to manage your conversation and processing.

    Examples
    --------
    >>> conversation = Conversation(recording=Path("/path/to/file.m4a"))
    >>> conversation.transcribe()
    >>> conversation.diarise()
    >>> html_report = conversation.report()
    """

    def __init__(
        self,
        recording: Path,
        num_speakers: int = 2,
        reload: bool = True,
        speaker_mapping: Optional[Dict[str, str]] = None,
        meeting_datetime: Optional[datetime] = None,
        attendees: Optional[List[str]] = None,
        transcription: Optional[Dict[str, str]] = None,
        transcription_shortened: Optional[str] = None,
        summary: Optional[str] = None,
        summary_automated: Optional[str] = None,
    ):
        """Initialise Conversations class.

        Parameters
        ----------
        recording : pathlib.Path
            Path to the conversation recording.
        num_speakers : int
            The number of speakers in the conversation.
        reload : bool
            If True, try to load an existing saved conversation with the default filename.
            If False, create a new Conversation instance.
        speaker_mapping : Optional[Dict[str, str]]
            A dictionary mapping speaker IDs to speaker names.

        Returns
        -------
        conversation : Conversation
            Instance of Conversation.
        """
        self._recording = recording
        self._num_speakers = num_speakers
        self._transcription = transcription
        self._diarisation: Optional[List[Dict[str, Any]]] = None
        self._speaker_mapping = speaker_mapping
        self._summary = summary
        self._summary_automated = summary_automated
        self._transcription_shortened = transcription_shortened

        if meeting_datetime is not None:
            self._meeting_datetime = meeting_datetime
        else:
            self._meeting_datetime = self._extract_datetime_from_file(recording)

        if reload:
            default_file_path = str(_create_default_save_filename(recording))
            try:
                loaded_conversation = load_conversation(default_file_path)
                self.__dict__.update(loaded_conversation.__dict__)
            except FileNotFoundError:
                print(f"No saved conversation was found at {default_file_path}.")

            # Overwrite reloaded values if user redefines them
            if speaker_mapping is not None:
                self._speaker_mapping = speaker_mapping

            if meeting_datetime is not None:
                self._meeting_datetime = meeting_datetime

        formatted_datetime = self._meeting_datetime.strftime("%Y-%m-%d %H:%M:%S %Z")
        print(f"Loaded conversation from: {formatted_datetime}")

        self._attendees = attendees  # Store the attendees
        if self._attendees is not None:
            print("Attendees:")
            for attendee in self._attendees:
                print(f"- {attendee}")

        # Set num_speakers based on attendees if not provided
        if num_speakers is None and attendees is not None:
            self._num_speakers = len(attendees)
        else:
            self._num_speakers = num_speakers

        # Verify that num_speakers is equal to the number of attendees
        if num_speakers is not None and attendees is not None:
            if num_speakers != len(attendees):
                raise ValueError(
                    "The number of speakers must match the number of attendees."
                )

    @staticmethod
    def _extract_datetime_from_file(file_path: Path) -> datetime:
        created_timestamp = os.path.getmtime(file_path)
        return datetime.fromtimestamp(created_timestamp, timezone.utc)

    def transcribe(self, method: str = "whisper", model: str = "medium.en"):
        """Transcribe a conversation."""
        from .transcribe import whisper

        if self._transcription is not None:
            print(
                "The conversation has already been transcribed. Skipping transcription."
            )
            return

        self._transcription = whisper.process(
            audio_file=self._recording, model_name=model
        )

    def diarise(self, method: str = "simple"):
        """Diarise a conversation."""
        from .diarise import simple

        if self._diarisation is not None:
            print("The conversation has already been diarised. Skipping diarisaton.")
            return

        self._diarisation = simple.process(
            audio_file=self._recording, num_speakers=self._num_speakers
        )

    def report(self, audio_file=None):
        """Generate a report of a conversation."""
        from .report import generate

        if audio_file is None:
            audio_file = self._recording

        return generate(
            transcript=self._transcription,
            audio_file=audio_file,
            diarisation=self._diarisation,
            speaker_mapping=self._speaker_mapping,
        )

    def export_text(self):
        """Export the transcript and diarisation as text."""
        from .report import export_text

        return export_text(
            transcript=self._transcription,
            diarisation=self._diarisation,
            speaker_mapping=self._speaker_mapping,
            datetimestr=self._meeting_datetime.strftime("%Y-%m-%d %H:%M:%S %Z"),
            attendees=self._attendees,
        )

    def save(self, file_path: Optional[str] = None) -> None:
        """Save the Conversation object to disk.

        Parameters
        ----------
        file_path : Optional[str]
            The path where the Conversation object will be saved.
            If None, a default filename based on the audio file path will be used.
        """
        if file_path is None:
            file_path = str(_create_default_save_filename(self._recording))

        with open(file_path, "wb") as file:
            pickle.dump(self, file)

    def summarise(
        self,
        force: bool = False,
        print_summary: bool = True,
        system_prompt: Optional[str] = None,
        summary_prompt: Optional[str] = None,
        append_prompt: Optional[str] = None,
    ):
        """Generate a summary of the conversation.

        Parameters
        ----------
        force : bool
            If True, generate a new summary even if one already exists.
        print_summary : bool
            If True, print the summary to the console.
        system_prompt : str or None
            The system prompt to use when generating the summary.
        summary_prompt : str or None
            The summary prompt to use when generating the summary.
        append_prompt : str or None
            The append prompt to use when generating the summary.

        Returns
        -------
        summary : str
            The generated summary.
        """
        if self._transcription_shortened is None:
            from .ai._chatgpt import _shorten_transcript

            self._transcription_shortened = _shorten_transcript(self.export_text())

        if (self._summary_automated is None) or force:
            from .ai import summarise

            self._summary_automated = summarise(
                self._transcription_shortened,
                system_prompt,
                summary_prompt,
                append_prompt,
            )

        if print_summary:
            print(self._summary_automated)

        return self._summary_automated

    def query(
        self,
        query: str,
        print_summary: bool = True,
        system_prompt: Optional[str] = None,
        append_prompt: Optional[str] = None,
    ):
        """Query the conversation.

        Parameters
        ----------
        query : str
            The query to ask the conversation.
        print_summary : bool
            If True, print the summary to the console.
        system_prompt : str or None
            The system prompt to use when generating the summary.
        append_prompt : str or None
            The append prompt to use when generating the summary.

        Returns
        -------
        summary : str
            The generated query response.
        """
        from .ai import query as query_fn

        if self._transcription_shortened is None:
            from .ai._chatgpt import _shorten_transcript

            self._transcription_shortened = _shorten_transcript(self.export_text())

        answer = query_fn(
            self._transcription_shortened, query, system_prompt, append_prompt
        )
        if print_summary:
            print(answer)
        return answer


def load_conversation(file_path: str) -> "Conversation":
    """Load a Conversation object from disk.

    Parameters
    ----------
    file_path : str
        The path to the file containing the saved Conversation object.

    Returns
    -------
    conversation : Conversation
        The loaded Conversation object.

    Raises
    ------
    ValueError
        If the loaded object is not an instance of the Conversation class.
    """
    with open(file_path, "rb") as file:
        conversation = pickle.load(file)

    if not isinstance(conversation, Conversation):
        raise ValueError(
            "The loaded object is not an instance of the Conversation class."
        )

    return conversation


def _create_default_save_filename(audio_file_path: Path) -> Path:
    """
    Create a default filename for saving conversations.

    Parameters
    ----------
    audio_file_path : pathlib.Path
        The path to the audio file.

    Returns
    -------
    pathlib.Path
        The default filename for saving the conversation.
    """
    file_stem = audio_file_path.stem
    default_filename = file_stem + ".conv"
    conversation_file_path = audio_file_path.with_name(default_filename)
    return conversation_file_path



from typing import Any, Dict, List, Optional
from ._utils import (
    _map_speaker_names,
    _get_segement_speaker,
    _group_transcript_segements_by_speaker,
)


def export_text(
    transcript: dict,
    diarisation: Optional[List[Dict[str, Any]]] = None,
    speaker_mapping: Optional[dict] = None,
    datetimestr: Optional[str] = None,
    attendees: Optional[List[str]] = None,
):
    """Create text file from conversation.

    Exports the transcript as a text document while grouping segments by speaker (if diarisation is provided) and
    mapping speaker names (if speaker_mapping is provided).

    Parameters
    ----------
    transcript : Dict[str, Any]
        The transcript data structured as a dictionary with a key "segments" containing a list of dicts.
        Each dict has a "text" key with the corresponding transcript segment text and optionally a "speaker" key.
    diarisation : Optional[List[Dict[str, Any]]], default=None
        A list of diarisation data dictionaries. Each dictionary should have the keys "start", "end", and "speaker".
    speaker_mapping : Optional[Dict[str, str]], default=None
        A dictionary that maps speaker identifiers to speaker names. The key is the speaker identifier and the value
        is the speaker name.
    datetimestr : Optional[str], default=None
        A string representing the date and time of the meeting.
    attendees : Optional[List[str]], default=None
        A list of attendees in the meeting.

    Returns
    -------
    str
        The transcript exported as a text document.

    Examples
    --------
    >>> transcript = {"segments": [{"text": "Hello, world!", "speaker": "speaker_1"}]}
    >>> diarisation = None
    >>> speaker_mapping = {"speaker_1": "John"}
    >>> export_text(transcript, diarisation, speaker_mapping)
    'John: Hello, world!'
    """
    if diarisation is not None:
        transcript = _group_transcript_segements_by_speaker(transcript, diarisation)

    if speaker_mapping is not None:
        transcript = _map_speaker_names(transcript, speaker_mapping)

    doc = ""

    previous_speaker = None

    if datetimestr is not None:
        doc += "Meeting Transcript from " + datetimestr
    if attendees is not None:
        doc += " with " + ", ".join(attendees)
    print("\n\n")

    for seg in transcript["segments"]:
        speaker = _get_segement_speaker(seg)

        if diarisation is not None:
            if speaker != previous_speaker:
                doc += f"\n\n{speaker}:"
                previous_speaker = speaker

        doc += f" {seg['text']}"

    return doc



from pathlib import Path
from simple_diarizer import diarizer
from typing import Any, Dict, List, Optional


def process(audio_file: Path, num_speakers: int = 2) -> List[Dict[str, Any]]:
    """Diarise audio using simple_diarizer.

    Parameters
    ----------
    audio_file : Path
        Path to the audio file.
    num_speakers : int
        The number of speakers in the conversation.

    Returns
    -------
    segments : list[dict]
        List containing the segments as dictionaries.
    """
    diar = diarizer.Diarizer(
        embed_model="xvec",  # 'xvec' and 'ecapa' supported
        cluster_method="sc",  # 'ahc' and 'sc' supported
    )

    segments = diar.diarize(audio_file, num_speakers=num_speakers)

    return segments



from conversations.transcribe import whisper
from pathlib import Path
from typing import Dict
import pooch


audio_file = pooch.retrieve(
    url="https://project-test-data-public.s3.amazonaws.com/test_audio.m4a",
    known_hash="md5:77d8b60c54dffbb74d48c4a65cd59591",
)


def test_local_process():
    """Test whisper processing of audio."""
    result = whisper.process(audio_file=Path(audio_file))
    assert isinstance(result, Dict)
    for seg in result["segments"]:
        assert "start" in seg
        assert "end" in seg
        assert "text" in seg


def test_cloud_process():
    """Test whisper processing of audio via api."""
    result = whisper.process(audio_file=Path(audio_file), model_name="openai.en")
    assert isinstance(result, Dict)
    for seg in result["segments"]:
        assert "start" in seg
        assert "end" in seg
        assert "text" in seg


def test_match_local_and_cloud_process():
    """Test that local and cloud processing return the same results."""
    local_result = whisper.process(audio_file=Path(audio_file))
    cloud_result = whisper.process(audio_file=Path(audio_file), model_name="openai.en")
    assert "million" in local_result["segments"][0]["text"]
    assert "million" in cloud_result["segments"][0]["text"]



from pathlib import Path
from typing import Dict

import openai
import whisper


def process(audio_file: Path, model_name: str = "base.en") -> Dict[str, str]:
    """Transcribe audio using Whisper.

    Parameters
    ----------
    audio_file : Path
        Path to the audio file.
    model_name : str, optional
        Name of the whisper model to use. To use the cloud service
        provided by OpenAI, use "openai.en", by default "base.en".

    Returns
    -------
    transcript : Dict[str, str]
        Dictionary containing the audio transcript in whisper format.
    """
    if model_name == "openai.en":
        result = _cloud_whisper(audio_file)
    else:
        result = _local_whisper(audio_file, model_name=model_name)
    return result


def _cloud_whisper(audio_file: Path) -> Dict[str, str]:
    """Transcribe audio using OpenAI's Whisper API.

    Parameters
    ----------
    audio_file : Path
        Path to the audio file.

    Returns
    -------
    transcript : Dict[str, str]
        Dictionary containing the audio transcript in whisper format.
    """
    with audio_file.open("rb") as audio:
        result = openai.Audio.transcribe(
            "whisper-1", audio, response_format="verbose_json"
        )
    return result


def _local_whisper(
    audio_file: Path,
    model_name: str = "base.en",
    device: str = "cpu",
) -> Dict[str, str]:
    """Transcribe audio using a local Whisper model.

    Parameters
    ----------
    audio_file : Path
        Path to the audio file.
    model_name : str, optional
        Name of the whisper model to use, available models are "base.en"
        and "large.en", by default "base.en".
    device : str, optional
        The device on which to run the model, either "cpu" or "gpu", by default "cpu".

    Returns
    -------
    transcript : Dict[str, str]
        Dictionary containing the audio transcript in whisper format.
    """
    model = whisper.load_model(model_name, device=device)
    audio = whisper.load_audio(str(audio_file))
    result = model.transcribe(audio)
    return result



"""ChatGPT based AI for conversations."""

from typing import List, Optional
import openai
import tiktoken


class ChunkTooLongError(Exception):
    """Raised when a chunk is too long for the model."""

    pass


def _num_tokens_from_messages(messages, model="gpt-4"):
    """Return the number of tokens used by a list of messages.

    Parameters
    ----------
    messages : List[dict]
        A list of message dictionaries.
    model : str, optional
        The model to use for tokenization, defaults to "gpt-4".

    Returns
    -------
    int
        The total number of tokens used by the messages.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo":
        print(
            "Warning: gpt-3.5-turbo may change over time. Returning num tokens assuming gpt-3.5-turbo-0301."
        )
        return _num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301")
    elif model == "gpt-4":
        print(
            "Warning: gpt-4 may change over time. Returning num tokens assuming gpt-4-0314."
        )
        return _num_tokens_from_messages(messages, model="gpt-4-0314")
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif model == "gpt-4-0314":
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def _chunk_transcript(transcript: str, chunk_size: int) -> List[str]:
    return [
        transcript[i : i + chunk_size] for i in range(0, len(transcript), chunk_size)
    ]


def _sumarise_chunk(chunk: str, model: str, temperature: float) -> str:
    system_prompt = (
        "You are an AI language model tasked with summarizing the following transcript "
        "while preserving important information. Please provide a concise summary of the text."
    )
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": chunk},
    ]

    response = openai.ChatCompletion.create(
        model=model, temperature=temperature, messages=messages
    )
    return response["choices"][0]["message"]["content"]


def _content_to_message(content: str):
    messages = [
        {"role": "system", "content": _system_prompt_summariser()},
        {
            "role": "user",
            "content": f"Reduce the length of this transcript to approximately half the length of the original text but not to remove any important information.\n\n {content}",
        },
    ]
    return messages


def _shorten_transcript(transcript: str, chunk_size: int = 8192) -> str:
    chunks = _chunk_transcript(transcript, chunk_size)

    summarized_chunks = []

    for chunk in chunks:
        messages = _content_to_message(chunk)

        if _num_tokens_from_messages(messages, model="gpt-4") > 8192:
            raise ChunkTooLongError(
                "The chunk is too long, consider reducing chunk_size."
            )

        summarized_chunk = _sumarise_chunk(chunk, model="gpt-4", temperature=0.7)
        summarized_chunks.append(summarized_chunk)

    shortened_transcript = " ".join(summarized_chunks).strip()

    if (
        _num_tokens_from_messages(
            _content_to_message(shortened_transcript), model="gpt-4"
        )
        > 8192
    ):
        print("Shortened transcript is still too long, reducing chunk size...")
        return _shorten_transcript(transcript, chunk_size=chunk_size)

    return shortened_transcript


def _system_prompt_summariser():
    """Generate the system prompt for the GPT model.

    Returns
    -------
    str
        The system prompt text.
    """
    system_prompt = """
    You are a helpful writing assistant. You write accurate and precise content.
    You will take the context from conversation notes that I provide and help me write useful summaries.
    You have excellent written communication skills and always use correct grammar and spelling.

    You are excellent at communication and will also over-communicate if you are unsure about an answer or need more information.
    You always chose to provide more information where possible.
    """
    return system_prompt


def _system_prompt_query():
    """Generate the system prompt for the GPT model.

    Returns
    -------
    str
        The system prompt text.
    """
    system_prompt = """
    You are a helpful assistant. You write accurate and precise response to questions I provide.
    You will take the context from conversation notes that I provide and provide accurate response.
    If you do not know the answer, say "I do not know the answer to that question".
    You have excellent written communication skills and always use correct grammar and spelling.
    Wherever possible, you will provide a quote from the conversation transcript to support your answer
    """
    return system_prompt


def _user_prompt_summariser(transcript) -> str:
    """Generate the summary prompt for the GPT model.

    Parameters
    ----------
    transcript : str
        The meeting transcript.

    Returns
    -------
    str
        The summary prompt text.
    """
    summary_prompt = f"""
    Meeting transcript:

    {transcript}

    First, write a 3 sentence summary of the meeting.

    Then summarise the key points of this meeting and include a quote from the transcript for each point.
    Be verbose and extract as many key points as possible and ensure you adequately summarise the conversation.
    Extract at least 10 key points.

    For each point, use the format:
    1. Point
       Speaker: Relevant quote

    List all action items discussed in the meeting and who is responsible for each action item.

    Next, extract 20 key words and list them in order from most to least relevant to the conversation.

    Next, extract 10 concepts and themes and list them in order from most to least relevant to the conversation.

    What was the most unexpected aspect of the conversation?

    Where any concepts or topics explained in the transcript? If so, list each concept and its definition as explained in the call. Highlight if the concept was incorrectly described in the call.

    What was the tone of the conversation? List the tone of the conversation as a whole and the tone of each speaker.
    """
    return summary_prompt


def _user_prompt_query(transcript: str, query: str) -> str:
    """Create a prompt for querying a transcript.

    Parameters
    ----------
    transcript : str
        The meeting transcript.
    query : str
        The query to ask the system.

    Returns
    -------
    str
        The query prompt text.
    """
    query_prompt = f"""
    Below is a transcript from conversation. Please read the transcript and then answer the following question:
    
    Meeting transcript:

    {transcript}
    
    Now, please answer the following question:

    {query}
    """
    return query_prompt


def summarise(
    transcript: str,
    system_prompt: Optional[str] = None,
    summary_prompt: Optional[str] = None,
    append_prompt: Optional[
        str
    ] = "Format your response as text and do not use markdown.",
) -> str:
    """Generate a meeting summary using the GPT model based on the given meeting transcript.

    Parameters
    ----------
    transcript : str
        The meeting transcript.
    system_prompt : Optional[str], default=None
        The system prompt text. If not provided, it will be generated using the internal function _system_prompt_summariser().
    summary_prompt : Optional[str], default=None
        The summary prompt text. If not provided, it will be generated using the internal function _user_prompt_summariser(transcript).
    append_prompt : Optional[str]
        An additional prompt to append to the summary_prompt.

    Returns
    -------
    str
        The generated meeting summary.
    """
    if system_prompt is None:
        system_prompt = _system_prompt_summariser()

    if summary_prompt is None:
        summary_prompt = _user_prompt_summariser(transcript)

    if append_prompt is not None:
        summary_prompt += append_prompt

    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": summary_prompt},
    ]

    response = openai.ChatCompletion.create(
        model="gpt-4", temperature=0.7, messages=messages
    )
    return response["choices"][0]["message"]["content"]


def query(
    transcript: str,
    query: str,
    system_prompt: Optional[str] = None,
    append_prompt: Optional[
        str
    ] = "Format your response as text and do not use markdown.",
) -> str:
    """Generate a response to a query using the GPT model based on the given meeting transcript.

    Parameters
    ----------
    transcript : str
        The meeting transcript.
    query : str
        What you would like to know from the conversation. This will be used to generate the query prompt.
    system_prompt : Optional[str], default=None
        The system prompt text. If not provided, it will be generated using the internal function _system_prompt_summariser().
    append_prompt : Optional[str]
        An additional prompt to append to the query_prompt.

    Returns
    -------
    answer : str
        The generated response to the query.
    """
    if system_prompt is None:
        system_prompt = _system_prompt_summariser()

    query_prompt = _user_prompt_query(transcript, query)

    if append_prompt is not None:
        query_prompt += append_prompt

    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": query_prompt},
    ]

    response = openai.ChatCompletion.create(
        model="gpt-4", temperature=0.7, messages=messages
    )
    return response["choices"][0]["message"]["content"]



import numpy as np


def _map_speaker_names(transcript: dict, speaker_mapping: dict) -> dict:
    """Map old speaker names to new speaker names in transcript.

    Parameters
    ----------
    transcript : dict
        Transcript from transcribe module.
    speaker_mapping : dict
        Mapping of old speaker names to new speaker names.

    Returns
    -------
    mapped_transcript : dict
        Transcript with updated speaker names.
    """
    mapped_segments = []

    for seg in transcript["segments"]:
        mapped_seg = seg.copy()
        if "speaker" in seg and seg["speaker"] in speaker_mapping:
            mapped_seg["speaker"] = speaker_mapping[seg["speaker"]]
        mapped_segments.append(mapped_seg)

    return {"segments": mapped_segments}


def _get_segement_speaker(segment) -> str:
    """Get speaker label from segment."""
    if "speaker" in segment:
        return segment["speaker"]
    else:
        return "unknown"


def _seconds_to_formatted(seconds):
    """Format seconds for use in report."""
    res = np.divmod(seconds, 60)
    min = res[0]
    sec = round(res[1])
    return f"{int(min):02}:{int(sec):02}"


def _find_current_speaker(timeval, diarisation) -> str:
    """Find the speaker label associated with a time."""
    for seg_idx in range(len(diarisation) - 1):
        seg = diarisation[seg_idx]
        next_seg = diarisation[seg_idx + 1]
        if (seg["start"] < timeval) & (next_seg["start"] > timeval):
            return str(seg["label"])
    return str(diarisation[-1]["label"])


def _group_transcript_segements_by_speaker(
    transcript, diarization, seconds_per_segment=20
):
    """Group transcript segments by speaker."""
    speaker_segments = []
    current_speaker = ""
    for seg in transcript["segments"]:
        start = seg["start"]
        stop = seg["end"]
        mid_time = start + ((stop - start) / 2)

        speaker = _find_current_speaker(mid_time, diarization)

        if speaker != current_speaker:
            current_speaker = speaker
            speaker_segments.append(
                {"speaker": speaker, "text": seg["text"], "start": start, "end": stop}
            )
        else:
            if stop - speaker_segments[-1]["start"] > seconds_per_segment:
                speaker_segments.append(
                    {
                        "speaker": speaker,
                        "text": seg["text"],
                        "start": start,
                        "end": stop,
                    }
                )
            else:
                speaker_segments[-1]["text"] += " " + seg["text"]
                speaker_segments[-1]["end"] = stop

    return {"segments": speaker_segments}



from conversations import report
from conversations.transcribe import whisper
from conversations.diarise import simple
from pathlib import Path
from typing import Dict
import pooch
import dominate


audio_file = pooch.retrieve(
    url="https://project-test-data-public.s3.amazonaws.com/test_audio.m4a",
    known_hash="md5:77d8b60c54dffbb74d48c4a65cd59591",
)

transcript = whisper.process(audio_file=audio_file)
diarisation = simple.process(audio_file=audio_file, num_speakers=3)


def test_report_without_audiofile():
    """Test report generation."""
    html_report = report.generate(transcript=transcript)
    assert isinstance(html_report, dominate.document)

    with open("tests/output/test_wo_audio.html", "w") as f:
        f.write(html_report.render())


def test_report_with_audiofile():
    """Test report generation."""
    html_report = report.generate(transcript=transcript, audio_file=audio_file)
    assert isinstance(html_report, dominate.document)

    with open("tests/output/test_w_audio.html", "w") as f:
        f.write(html_report.render())


def test_report_with_mapping():
    """Test report generation."""
    html_report = report.generate(
        transcript=transcript,
        audio_file=audio_file,
        speaker_mapping={"0": "Alice", "1": "Bob", "2": "Sam"},
    )
    assert isinstance(html_report, dominate.document)

    with open("tests/output/test_w_audio.html", "w") as f:
        f.write(html_report.render())


def test_text_export():
    """Test report generation."""
    text_report = report.export_text(
        transcript=transcript,
        speaker_mapping={"0": "Alice", "1": "Bob", "2": "Sam"},
        diarisation=diarisation,
    )
    assert isinstance(text_report, str)

    with open("conversation.txt", "w") as f:
        f.write(text_report)
